"""
A major part of this code is translated from https://github.com/facebookresearch/mae
"""

import jax.numpy as jnp

import flax.linen as nn
from ..layers import Attention, DropPath, TransformerMLP, Mask, PatchEmbed
from typing import Optional
import logging

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)

__all__ = ["MaskedAutoencoderViT", "MAE_Base", "MAE_Large", "MAE_Huge"]


def sincosemb1d(emb_dim, pos):
    omega = jnp.arange(emb_dim // 2, dtype=jnp.float32)
    omega /= emb_dim / 2.0
    omega = 1.0 / 10000 * omega

    pos = pos.reshape(-1)
    out = jnp.einsum("m, d -> md", pos, omega)

    return jnp.concatenate([jnp.sin(out), jnp.cos(out)], axis=1)


def sincosposemb(emb_dim, grid_size):
    grid_h = jnp.arange(grid_size, dtype=jnp.float32)
    grid_w = jnp.arange(grid_size, dtype=jnp.float32)
    grid = jnp.meshgrid(grid_w, grid_h)
    grid = jnp.stack(grid, axis=0)
    grid = jnp.reshape(grid, [2, 1, grid_size, grid_size])

    emb_height = sincosemb1d(emb_dim // 2, grid[0])
    emb_width = sincosemb1d(emb_dim // 2, grid[1])

    pos_emb = jnp.concatenate([emb_height, emb_width], axis=1)
    return jnp.expand_dims(jnp.concatenate([jnp.zeros([1, emb_dim]), pos_emb], 0), 0)


pos_emb_init = sincosposemb
emb_init = nn.initializers.normal(stddev=0.02)


class Block(nn.Module):
    dim: int = 256
    num_heads: int = 8
    mlp_ratio: int = 4
    att_drop: float = 0.1
    drop: float = 0.1
    drop_path: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        x = nn.LayerNorm()(inputs)
        x = Attention(self.dim, self.num_heads, True, self.att_drop, self.drop)(
            x, deterministic
        )
        x = DropPath(self.drop_path)(x, deterministic)
        inputs = inputs + x

        x = nn.LayerNorm()(inputs)
        x = TransformerMLP(
            self.dim * self.mlp_ratio, self.dim, self.drop, use_dwconv=False
        )(x, deterministic)
        x = DropPath(self.drop_path)(x, deterministic)
        return inputs + x


class Encoder(nn.Module):
    patch_size: int = 16
    emb_dim: int = 1024
    depth: int = 24
    mask_ratio: float = 0.5
    att_drop: float = 0
    drop: float = 0
    drop_path: float = 0
    num_heads: int = 16
    mlp_ratio: int = 4
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, patches, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        patch_embed = PatchEmbed(self.patch_size, self.emb_dim, use_norm=False)(patches)
        num_patches = patch_embed.shape[1]

        pos_emb = self.variable(
            "pos_emb",
            "enc_pos_emb",
            pos_emb_init,
            self.emb_dim,
            int(num_patches ** 0.5),
        )

        x = patch_embed + pos_emb.value[:, 1:, :]
        x, mask, ids_restore = Mask()(x, self.mask_ratio)

        cls_token = self.param("cls_token", emb_init, (1, 1, self.emb_dim))
        cls_token = cls_token + pos_emb.value[:, :1, :]
        cls_token = jnp.broadcast_to(
            cls_token, (x.shape[0], cls_token.shape[1], cls_token.shape[2])
        )
        x = jnp.concatenate([cls_token, x], axis=1)

        for _ in range(self.depth):
            x = Block(
                self.emb_dim,
                self.num_heads,
                self.mlp_ratio,
                self.att_drop,
                self.drop,
                self.drop_path,
            )(x, deterministic)

        x = nn.LayerNorm()(x)

        return x, mask, ids_restore, num_patches


class Decoder(nn.Module):
    patch_size: int = 16
    dec_emb_dim: int = 512
    dec_depth: int = 8
    num_patches: int = 196
    att_drop: float = 0
    drop: float = 0
    drop_path: float = 0
    dec_num_heads: int = 16
    mlp_ratio: int = 4
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, encoder_output, ids_restore, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        x = nn.Dense(self.dec_emb_dim)(encoder_output)

        mask_token = self.param("mask_token", emb_init, (1, 1, self.dec_emb_dim))
        mask_tokens = jnp.tile(
            mask_token, (x.shape[0], ids_restore.shape[1] + 1 - x.shape[1], 1)
        )

        x_ = jnp.concatenate([x[:, 1:, :], mask_tokens], axis=1)
        x_ = jnp.take_along_axis(
            x_, jnp.tile(jnp.expand_dims(ids_restore, -1), (1, 1, x.shape[2])), axis=1
        )
        x = jnp.concatenate([x[:, :1, :], x_], axis=1)

        decoder_pos_emb = self.variable(
            "pos_emb",
            "dec_pos_emb",
            pos_emb_init,
            self.dec_emb_dim,
            int(self.num_patches ** 0.5),
        )

        x = x + decoder_pos_emb.value

        for _ in range(self.dec_depth):
            x = Block(
                self.dec_emb_dim,
                self.dec_num_heads,
                self.mlp_ratio,
                self.att_drop,
                self.drop,
                self.drop_path,
            )(x, deterministic)

        x = nn.LayerNorm()(x)
        x = nn.Dense(self.patch_size ** 2 * 3)(x)

        return x[:, 1:, :]


class MaskedAutoencoderViT(nn.Module):
    patch_size: int = 16
    emb_dim: int = 1024
    dec_emb_dim: int = 512
    depth: int = 24
    dec_depth: int = 8
    mask_ratio: float = 0.5
    att_drop: float = 0
    drop: float = 0
    drop_path: float = 0
    num_heads: int = 16
    dec_num_heads: int = 16
    mlp_ratio: int = 4
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        x, mask, ids_restore, num_patches = Encoder(
            self.patch_size,
            self.emb_dim,
            self.depth,
            self.mask_ratio,
            self.att_drop,
            self.drop,
            self.drop_path,
            self.num_heads,
            self.mlp_ratio,
        )(inputs, deterministic)

        x = Decoder(
            self.patch_size,
            self.dec_emb_dim,
            self.dec_depth,
            num_patches,
            self.att_drop,
            self.drop,
            self.drop_path,
            self.dec_num_heads,
            self.mlp_ratio,
        )(x, ids_restore, deterministic)

        return x, mask


def MAE_Base(attach_head=None, num_classes=None, dropout=None, **kwargs):
    logging.info(
        "Default classification arguments are ignored since this is a generative model. To tune the hyperparameters, please call `MaskedAutoencoderViT` separately with your desired arguments."
    )
    del attach_head, num_classes, dropout
    return MaskedAutoencoderViT(
        patch_size=16,
        emb_dim=768,
        depth=12,
        num_heads=12,
        dec_emb_dim=512,
        dec_depth=8,
        dec_num_heads=16,
        mlp_ratio=4,
        **kwargs
    )


def MAE_Large(attach_head=None, num_classes=None, dropout=None, **kwargs):
    logging.info(
        "Default classification arguments are ignored since this is a generative model. To tune the hyperparameters, please call `MaskedAutoencoderViT` separately with your desired arguments."
    )
    del attach_head, num_classes, dropout
    return MaskedAutoencoderViT(
        patch_size=16,
        emb_dim=1024,
        depth=24,
        num_heads=16,
        dec_emb_dim=512,
        dec_depth=8,
        dec_num_heads=16,
        mlp_ratio=4,
        **kwargs
    )


def MAE_Huge(attach_head=None, num_classes=None, dropout=None, **kwargs):
    logging.info(
        "Default classification arguments are ignored since this is a generative model. To tune the hyperparameters, please call `MaskedAutoencoderViT` separately with your desired arguments."
    )
    del attach_head, num_classes, dropout
    return MaskedAutoencoderViT(
        patch_size=14,
        emb_dim=1280,
        depth=32,
        num_heads=16,
        dec_emb_dim=512,
        dec_depth=8,
        dec_num_heads=16,
        mlp_ratio=4,
        **kwargs
    )



import jax.random as random
import jax.numpy as jnp
import flax.linen as nn
from typing import Optional, Iterable, Sequence

from ..layers import DepthwiseConv2D, DropPath

__all__ = [
    "ConvNeXt",
    "ConvNeXt_Tiny",
    "ConvNeXt_Small",
    "ConvNeXt_Base",
    "ConvNeXt_Large",
    "ConvNeXt_XLarge",
]

initializer = nn.initializers.variance_scaling(
    0.2, "fan_in", distribution="truncated_normal"
)


class ConvNeXtBlock(nn.Module):
    dim: int = 256
    layer_scale_init_value: float = 1e-6
    drop_path: float = 0.1
    deterministic: Optional[bool] = None

    def init_fn(self, key, shape, fill_value):
        return jnp.full(shape, fill_value)

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        x = DepthwiseConv2D((7, 7), weights_init=initializer)(inputs)
        x = nn.LayerNorm()(x)
        x = nn.Dense(4 * self.dim, kernel_init=initializer)(x)
        x = nn.gelu(x)
        x = nn.Dense(self.dim, kernel_init=initializer)(x)
        if self.layer_scale_init_value > 0:
            gamma = self.param(
                "gamma", self.init_fn, (self.dim,), self.layer_scale_init_value
            )
            x = gamma * x

        x = inputs + DropPath(self.drop_path)(x, deterministic)
        return x


class ConvNeXt(nn.Module):
    depths: Iterable = (3, 3, 9, 3)
    dims: Iterable = (96, 192, 384, 768)
    drop_path: float = 0.1
    layer_scale_init_value: float = 1e-6
    head_init_scale: float = 1.0
    attach_head: bool = False
    num_classes: int = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        dp_rates = [x.item() for x in jnp.linspace(0, self.drop_path, sum(self.depths))]
        curr = 0

        # Stem
        x = nn.Conv(self.dims[0], (4, 4), 4, kernel_init=initializer)(inputs)
        x = nn.LayerNorm()(x)
        for j in range(self.depths[0]):
            x = ConvNeXtBlock(
                self.dims[0],
                drop_path=dp_rates[curr + j],
                layer_scale_init_value=self.layer_scale_init_value,
            )(x, deterministic)
        curr += self.depths[0]

        # Downsample layers
        for i in range(3):
            x = nn.LayerNorm()(x)
            x = nn.Conv(self.dims[i + 1], (2, 2), 2, kernel_init=initializer)(x)

            for j in range(self.depths[i + 1]):
                x = ConvNeXtBlock(
                    self.dims[i + 1],
                    drop_path=dp_rates[curr + j],
                    layer_scale_init_value=self.layer_scale_init_value,
                )(x, deterministic)

            curr += self.depths[i + 1]

        if self.attach_head:
            x = nn.LayerNorm()(x)
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes, kernel_init=initializer)(x)
        return x


def ConvNeXt_Tiny(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return ConvNeXt(
        depths=(3, 3, 9, 3),
        dims=(96, 192, 384, 768),
        drop_path=dropout,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )


def ConvNeXt_Small(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return ConvNeXt(
        depths=(3, 3, 27, 3),
        dims=(96, 192, 384, 768),
        drop_path=dropout,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )


def ConvNeXt_Base(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return ConvNeXt(
        depths=(3, 3, 27, 3),
        dims=(128, 256, 512, 1024),
        drop_path=dropout,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )


def ConvNeXt_Large(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return ConvNeXt(
        depths=(3, 3, 27, 3),
        dims=(192, 384, 768, 1536),
        drop_path=dropout,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )


def ConvNeXt_XLarge(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return ConvNeXt(
        depths=(3, 3, 27, 3),
        dims=(256, 512, 1024, 2048),
        drop_path=dropout,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )



from typing import Callable
import jax.numpy as jnp
import flax.linen as nn


class ExtractPatches(nn.Module):
    @nn.compact
    def __call__(self, inputs, patch_size):
        batch, height, width, channels = inputs.shape
        height, width = height // patch_size, width // patch_size
        x = jnp.reshape(
            inputs, (batch, height, patch_size, width, patch_size, channels)
        )
        x = jnp.reshape(x, (batch, height * width, patch_size ** 2 * channels))
        return x


class MergePatches(nn.Module):
    @nn.compact
    def __call__(self, inputs, patch_size):
        batch, length, _ = inputs.shape
        height = width = int(length ** 0.5)
        x = jnp.reshape(inputs, (batch, height, patch_size, width, patch_size, -1))
        x = jnp.reshape(x, (batch, height * patch_size, width * patch_size, -1))
        return x


class PatchEmbed(nn.Module):
    patch_size: int = 16
    emb_dim: int = 768
    use_norm: bool = False
    kernel_init: Callable = nn.initializers.xavier_uniform()

    @nn.compact
    def __call__(self, inputs):
        batch, height, width, channels = inputs.shape
        x = nn.Conv(
            self.emb_dim,
            (self.patch_size, self.patch_size),
            self.patch_size,
            kernel_init=self.kernel_init,
        )(inputs)
        x = jnp.reshape(x, (batch, -1, self.emb_dim))
        if self.use_norm:
            x = nn.LayerNorm()(x)
        return x



import flax.linen as nn
import jax.numpy as jnp

from typing import Optional
from ..layers import TransformerMLP

__all__ = [
    "PoolFormer_S12",
    "PoolFormer_S24",
    "PoolFormer_S36",
    "PoolFormer_M36",
    "PoolFormer_M48",
]


class TransformerEncoder(nn.Module):
    mlp_dim: int
    pool_size: int
    stride: int
    dropout: float
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        norm = nn.LayerNorm()(inputs)
        att = nn.avg_pool(
            norm,
            (self.pool_size, self.pool_size),
            strides=(self.stride, self.stride),
            padding="SAME",
        )
        att = att - norm
        add = inputs + att
        x = nn.LayerNorm()(add)
        x = TransformerMLP(self.mlp_dim, self.mlp_dim, self.dropout)(x, deterministic)
        return add + x


class AddPositionEmbs(nn.Module):
    @nn.compact
    def __call__(self, inputs):
        pos_emb_shape = (1, inputs.shape[1], inputs.shape[2], inputs.shape[3])
        pe = self.param(
            "pos_embedding", nn.initializers.normal(stddev=0.02), pos_emb_shape
        )
        return inputs + pe


class S12(nn.Module):
    attach_head: bool = False
    num_classes: int = 1000
    dropout: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        # H/4, W/4
        x = nn.Conv(64, (7, 7), 4, "SAME")(inputs)
        x = AddPositionEmbs()(x)
        for _ in range(2):
            x = TransformerEncoder(64, 3, 1, self.dropout)(x, deterministic)

        # H/8, W/8
        x = nn.Conv(128, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(2):
            x = TransformerEncoder(128, 3, 1, self.dropout)(x, deterministic)

        # H/16, W/16
        x = nn.Conv(320, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(320, 3, 1, self.dropout)(x, deterministic)

        # H/32, W/32
        x = nn.Conv(512, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(2):
            x = TransformerEncoder(512, 3, 1, self.dropout)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)
        return x


class S24(nn.Module):
    attach_head: bool = False
    num_classes: int = 1000
    dropout: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        # H/4, W/4
        x = nn.Conv(64, (7, 7), 4, "SAME")(inputs)
        x = AddPositionEmbs()(x)
        for _ in range(4):
            x = TransformerEncoder(64, 3, 1, self.dropout)(x, deterministic)

        # H/8, W/8
        x = nn.Conv(128, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(4):
            x = TransformerEncoder(128, 3, 1, self.dropout)(x, deterministic)

        # H/16, W/16
        x = nn.Conv(320, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(12):
            x = TransformerEncoder(320, 3, 1, self.dropout)(x, deterministic)

        # H/32, W/32
        x = nn.Conv(512, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(4):
            x = TransformerEncoder(512, 3, 1, self.dropout)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


class S36(nn.Module):
    attach_head: bool = False
    num_classes: int = 1000
    dropout: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        # H/4, W/4
        x = nn.Conv(64, (7, 7), 4, "SAME")(inputs)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(64, 3, 1, self.dropout)(x, deterministic)

        # H/8, W/8
        x = nn.Conv(128, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(128, 3, 1, self.dropout)(x, deterministic)

        # H/16, W/16
        x = nn.Conv(320, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(18):
            x = TransformerEncoder(320, 3, 1, self.dropout)(x, deterministic)

        # H/32, W/32
        x = nn.Conv(512, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(512, 3, 1, self.dropout)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


class M36(nn.Module):
    attach_head: bool = False
    num_classes: int = 1000
    dropout: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        # H/4, W/4
        x = nn.Conv(96, (7, 7), 4, "SAME")(inputs)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(96, 3, 1, self.dropout)(x, deterministic)

        # H/8, W/8
        x = nn.Conv(192, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(192, 3, 1, self.dropout)(x, deterministic)

        # H/16, W/16
        x = nn.Conv(384, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(18):
            x = TransformerEncoder(384, 3, 1, self.dropout)(x, deterministic)

        # H/32, W/32
        x = nn.Conv(768, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(6):
            x = TransformerEncoder(768, 3, 1, self.dropout)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


class M48(nn.Module):
    attach_head: bool = False
    num_classes: int = 1000
    dropout: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        # H/4, W/4
        x = nn.Conv(96, (7, 7), 4, "SAME")(inputs)
        x = AddPositionEmbs()(x)
        for _ in range(8):
            x = TransformerEncoder(96, 3, 1, self.dropout)(x, deterministic)

        # H/8, W/8
        x = nn.Conv(192, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(8):
            x = TransformerEncoder(192, 3, 1, self.dropout)(x, deterministic)

        # H/16, W/16
        x = nn.Conv(384, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(24):
            x = TransformerEncoder(384, 3, 1, self.dropout)(x, deterministic)

        # H/32, W/32
        x = nn.Conv(768, (3, 3), 2, "SAME")(x)
        x = AddPositionEmbs()(x)
        for _ in range(8):
            x = TransformerEncoder(768, 3, 1, self.dropout)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


def PoolFormer_S12(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return S12(attach_head, num_classes, dropout, **kwargs)


def PoolFormer_S24(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return S24(attach_head, num_classes, dropout, **kwargs)


def PoolFormer_S36(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return S36(attach_head, num_classes, dropout, **kwargs)


def PoolFormer_M36(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return M36(attach_head, num_classes, dropout, **kwargs)


def PoolFormer_M48(attach_head=False, num_classes=1000, dropout=0.1, **kwargs):
    return M48(attach_head, num_classes, dropout, **kwargs)



import jax.numpy as jnp

import flax.linen as nn
from typing import Optional

from ..layers import SqueezeAndExcitation, DropPath, TransformerMLP

__all__ = [
    "PatchConvNet",
    "PatchConvNet_S60",
    "PatchConvNet_S120",
    "PatchConvNet_B60",
    "PatchConvNet_B120",
    "PatchConvNet_L60",
    "PatchConvNet_L120",
]


class ConvolutionalStem(nn.Module):
    emb_dim: int = 768

    @nn.compact
    def __call__(self, inputs):
        conv = nn.Conv(
            self.emb_dim // 8, (3, 3), strides=2, padding="SAME", use_bias=False
        )(inputs)
        conv = nn.gelu(conv)
        conv = nn.Conv(
            self.emb_dim // 4, (3, 3), strides=2, padding="SAME", use_bias=False
        )(conv)
        conv = nn.gelu(conv)
        conv = nn.Conv(
            self.emb_dim // 2, (3, 3), strides=2, padding="SAME", use_bias=False
        )(conv)
        conv = nn.gelu(conv)
        conv = nn.Conv(self.emb_dim, (3, 3), strides=2, padding="SAME", use_bias=False)(
            conv
        )
        batch, height, width, channels = conv.shape
        return jnp.reshape(conv, (batch, height * width, channels))


class TrunkBlock(nn.Module):
    conv_dim: int

    @nn.compact
    def __call__(self, inputs):
        batch, proj_dim, channels = inputs.shape
        height = width = int(proj_dim ** 0.5)
        reshaped = jnp.reshape(inputs, (batch, height, width, channels))

        norm = nn.LayerNorm()(reshaped)
        conv1 = nn.Conv(self.conv_dim, (1, 1), strides=1)(norm)
        conv1 = nn.gelu(conv1)
        conv2 = nn.Conv(self.conv_dim, (3, 3), strides=1)(conv1)
        conv2 = nn.gelu(conv2)
        sae = SqueezeAndExcitation()(conv2)
        conv3 = nn.Conv(self.conv_dim, (1, 1), strides=1)(sae)
        conv3 = jnp.reshape(conv3, (batch, proj_dim, channels))

        gamma = self.param("gamma", nn.ones, (self.conv_dim,))
        scaled_output = gamma * conv3
        return inputs + scaled_output


class Aggregation(nn.Module):
    dim: int = 768
    dropout: float = 0.5
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        batch, proj_dim, channels = inputs.shape
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        q = nn.Dense(self.dim, use_bias=False)(inputs[:, 0])
        q = jnp.transpose(
            jnp.expand_dims(q, 1).reshape(batch, 1, 1, channels), (0, 2, 1, 3)
        )

        k = nn.Dense(self.dim, use_bias=False)(inputs)
        k = jnp.reshape(k, (batch, proj_dim, 1, channels))
        k = jnp.transpose(k, (0, 2, 1, 3))

        q = q * (self.dim ** -0.5)
        v = nn.Dense(self.dim, use_bias=False)(inputs)
        v = jnp.transpose(jnp.reshape(v, (batch, proj_dim, 1, channels)), (0, 2, 1, 3))

        att = jnp.matmul(q, jnp.transpose(k, (0, 1, 3, 2)))
        att = nn.softmax(att, -1)
        att_drop = nn.Dropout(self.dropout)(att, deterministic=deterministic)

        x_cls = jnp.transpose(jnp.matmul(att_drop, v), (0, 2, 1, 3)).reshape(
            batch, 1, channels
        )
        x_cls = nn.Dense(self.dim)(x_cls)
        x_cls = nn.Dropout(self.dropout)(x_cls, deterministic=deterministic)

        return x_cls


class AttentionPoolingBlock(nn.Module):
    dim: int = 768
    dropout: float = 0.5
    mlp_ratio: int = 4
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        input, token = inputs
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        hidden_dim = int(self.dim * self.mlp_ratio)
        gamma_1 = self.param("gamma_1", nn.zeros, (self.dim,))
        gamma_2 = self.param("gamma_2", nn.zeros, (self.dim,))

        concat = jnp.concatenate((token, input), axis=1)

        norm = nn.LayerNorm()(concat)
        att = Aggregation(self.dim, self.dropout)(norm, deterministic=deterministic)
        scaled_att = gamma_1 * att
        drop_att = DropPath(self.dropout)(scaled_att, deterministic=deterministic)
        token = token + drop_att

        norm = nn.LayerNorm()(token)
        x = TransformerMLP(dim=hidden_dim, out_dim=self.dim, dropout=self.dropout)(
            norm, deterministic
        )
        scaled_mlp = gamma_2 * x
        drop_mlp = DropPath(self.dropout)(scaled_mlp, deterministic=deterministic)
        token = token + drop_mlp

        return token


class PatchConvNet(nn.Module):
    depth: int = 20
    dim: int = 768
    dropout: float = 0.5
    mlp_ratio: int = 4
    attach_head: bool = True
    num_classes: int = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        stem = ConvolutionalStem(emb_dim=self.dim)(inputs)
        cls_token = self.param(
            "cls_token", nn.zeros, (inputs.shape[0], 1, int(self.dim))
        )

        trunk = stem
        for _ in range(self.depth):
            trunk = TrunkBlock(conv_dim=self.dim)(trunk)

        cls_token = AttentionPoolingBlock(
            dim=self.dim, dropout=self.dropout, mlp_ratio=self.mlp_ratio
        )([trunk, cls_token], deterministic=deterministic)

        x = jnp.concatenate((cls_token, trunk), axis=1)
        x = nn.LayerNorm()(x)
        x = x[:, 0]
        if self.attach_head:
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)
        return x


def PatchConvNet_S60(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=60,
        dim=384,
        dropout=dropout,
        mlp_ratio=3,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model


def PatchConvNet_S120(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=120,
        dim=384,
        dropout=dropout,
        mlp_ratio=3,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model


def PatchConvNet_B60(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=60,
        dim=768,
        dropout=dropout,
        mlp_ratio=4,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model


def PatchConvNet_B120(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=120,
        dim=768,
        dropout=dropout,
        mlp_ratio=4,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model


def PatchConvNet_L60(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=60,
        dim=1024,
        dropout=dropout,
        mlp_ratio=3,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model


def PatchConvNet_L120(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    model = PatchConvNet(
        depth=120,
        dim=1024,
        dropout=dropout,
        mlp_ratio=3,
        attach_head=attach_head,
        num_classes=num_classes,
        **kwargs
    )
    return model



import jax.numpy as jnp
import flax.linen as nn
from typing import Union, Iterable, Callable


def relu6(input):
    return jnp.minimum(jnp.maximum(input, 0), 6)


def hardswish(input):
    return input * (relu6(input + 3) / 6)


def mish(input):
    return input * jnp.tanh(nn.softplus(input))


class PReLU(nn.Module):
    alpha_init: Callable = nn.zeros
    shared_axis: Union[Iterable, int] = None

    @nn.compact
    def __call__(self, inputs):
        param_shape = list(inputs.shape[1:])
        if self.shared_axis is not None:
            for i in self.shared_axes:
                param_shape[i - 1] = 1
        alpha = self.param("alpha", self.alpha_init, tuple(param_shape))
        pos = nn.relu(inputs)
        neg = -alpha * nn.relu(-inputs)
        return pos + neg



import jax
import jax.numpy as jnp

import flax.linen as nn
from typing import Optional, Union, Tuple, Callable, Sequence


class DepthwiseConv2D(nn.Module):
    kernel_shape: Union[int, Sequence[int]] = (1, 1)
    stride: Union[int, Sequence[int]] = (1, 1)
    padding: str or Sequence[Tuple[int, int]] = "SAME"
    channel_multiplier: int = 1
    use_bias: bool = True
    weights_init: Callable = nn.initializers.lecun_uniform()
    bias_init: Optional[Callable] = nn.initializers.zeros

    @nn.compact
    def __call__(self, input):
        w = self.param(
            "weights",
            self.weights_init,
            self.kernel_shape + (1, self.channel_multiplier * input.shape[-1]),
        )
        if self.use_bias:
            b = self.param(
                "bias", self.bias_init, (self.channel_multiplier * input.shape[-1],)
            )

        conv = jax.lax.conv_general_dilated(
            input,
            w,
            self.stride,
            self.padding,
            (1,) * len(self.kernel_shape),
            (1,) * len(self.kernel_shape),
            ("NHWC", "HWIO", "NHWC"),
            input.shape[-1],
        )
        if self.use_bias:
            bias = jnp.broadcast_to(b, conv.shape)
            return conv + bias
        else:
            return conv


class SeparableDepthwiseConv2D(nn.Module):
    """Separable 2-D Depthwise Convolution Module."""

    features: int
    channel_multiplier: int
    kernel_shape: Union[int, Sequence[int]]
    stride: Union[int, Sequence[int]] = 1
    padding: Union[str, Sequence[Tuple[int, int]]] = "SAME"
    use_bias: bool = True
    weights_init: Callable = nn.initializers.lecun_uniform()
    bias_init: Optional[Callable] = nn.initializers.zeros

    @nn.compact
    def __call__(self, inputs):
        depthwise_conv = DepthwiseConv2D(
            channel_multiplier=self.channel_multiplier,
            kernel_shape=self.kernel_shape,
            stride=self.stride,
            padding=self.padding,
            use_bias=False,
            weights_init=self.weights_init,
            bias_init=self.bias_init,
        )

        pointwise_conv = nn.Conv(
            self.features,
            kernel_size=[1, 1],
            strides=1,
            padding=self.padding,
            use_bias=False,
            kernel_init=self.weights_init,
            bias_init=self.bias_init,
        )

        sep_conv = pointwise_conv(depthwise_conv(inputs))

        if self.use_bias:
            b = self.param("bias", nn.zeros, (sep_conv.shape[-1]))

            return sep_conv + b

        return sep_conv



from .conv_mixer import *
from .mlp_mixer import *
from .mpvit import *
from .patchconvnet import *
from .poolformer import *
from .segformer import *
from .convnext import *
from .masked_autoencoder import *
from .swin_transformer import *


model_dict = {
    "mpvit-tiny": MPViT_Tiny,
    "mpvit-xsmall": MPViT_XSmall,
    "mpvit-small": MPViT_Small,
    "mpvit-base": MPViT_Base,
    "patchconvnet-s60": PatchConvNet_S60,
    "patchconvnet-s120": PatchConvNet_S120,
    "patchconvnet-b60": PatchConvNet_B60,
    "patchconvnet-b120": PatchConvNet_B120,
    "patchconvnet-l60": PatchConvNet_L60,
    "patchconvnet-l120": PatchConvNet_L120,
    "poolformer-s12": PoolFormer_S12,
    "poolformer-s24": PoolFormer_S24,
    "poolformer-s36": PoolFormer_S36,
    "poolformer-m36": PoolFormer_M36,
    "poolformer-m48": PoolFormer_M48,
    "mlpmixer-s16": MLPMixer_S16,
    "mlpmixer-s32": MLPMixer_S32,
    "mlpmixer-b32": MLPMixer_B32,
    "mlpmixer-l16": MLPMixer_L16,
    "mlpmixer-l32": MLPMixer_L32,
    "mlpmixer-h14": MLPMixer_H14,
    "convmixer-1536-20": ConvMixer_1536_20,
    "convmixer-1024-20": ConvMixer_1024_20,
    "convmixer-768-32": ConvMixer_768_32,
    "convmixer-512-12": ConvMixer_512_12,
    "segformer-b0": SegFormer_B0,
    "segformer-b1": SegFormer_B1,
    "segformer-b2": SegFormer_B2,
    "segformer-b3": SegFormer_B3,
    "segformer-b4": SegFormer_B4,
    "segformer-b5": SegFormer_B5,
    "convnext-tiny": ConvNeXt_Tiny,
    "convnext-small": ConvNeXt_Small,
    "convnext-base": ConvNeXt_Base,
    "convnext-large": ConvNeXt_Large,
    "convnext-xlarge": ConvNeXt_XLarge,
    "mae-base": MAE_Base,
    "mae-large": MAE_Large,
    "mae-huge": MAE_Huge,
    "swin-tiny-224": SwinTiny224,
    "swin-small-224": SwinSmall224,
    "swin-base-224": SwinBase224,
    "swin-base-384": SwinBase384,
    "swin-large-224": SwinLarge224,
    "swin-large-384": SwinLarge384,
}


def list_models():
    return sorted(list(model_dict.keys()))


def load_model(
    model_str="", attach_head=False, num_classes=1000, dropout=0.0, **kwargs
):
    return model_dict[model_str](attach_head, num_classes, dropout=dropout, **kwargs)



import jax.numpy as jnp

import flax.linen as nn


class SqueezeAndExcitation(nn.Module):
    reduction: int = 16

    @nn.compact
    def __call__(self, inputs):
        batch, _, _, channels = inputs.shape
        global_avg_pool = jnp.mean(inputs, axis=[1, 2], keepdims=False)
        dense1 = nn.Dense(channels // self.reduction, use_bias=False)(global_avg_pool)
        dense1 = nn.relu(dense1)
        dense2 = nn.Dense(channels, use_bias=False)(dense1)
        dense2 = nn.sigmoid(dense2)
        expand_dims = jnp.reshape(dense2, (batch, 1, 1, channels))
        shape_broadcasted = jnp.broadcast_to(expand_dims, inputs.shape)
        return inputs * shape_broadcasted



import jax.numpy as jnp
import flax.linen as nn

from ..layers import DepthwiseConv2D
from typing import Optional, Callable


class TransformerMLP(nn.Module):
    dim: int = 256
    out_dim: int = 256
    dropout: float = 0.2
    use_dwconv: bool = False
    kernel_init: Callable = nn.initializers.xavier_uniform()
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        x = nn.Dense(self.dim, kernel_init=self.kernel_init)(inputs)

        if self.use_dwconv:
            batch, n, channels = x.shape
            width = height = int(n ** 0.5)
            x = jnp.reshape(x, (batch, height, width, channels))
            x = DepthwiseConv2D((3, 3))(x)
            x = jnp.reshape(x, (batch, -1, channels))

        x = nn.gelu(x)
        x = nn.Dropout(self.dropout)(x, deterministic)
        x = nn.Dense(self.out_dim, kernel_init=self.kernel_init)(x)
        x = nn.Dropout(self.dropout)(x, deterministic)

        return x



import jax.numpy as jnp
import flax.linen as nn

from typing import Optional, Callable


class Attention(nn.Module):
    dim: int
    num_heads: int = 8
    use_bias: bool = False
    att_drop: float = 0
    proj_drop: float = 0
    kernel_init: Callable = nn.initializers.xavier_uniform()
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        batch, n, channels = inputs.shape
        scale = (self.dim // self.num_heads) ** -0.5
        qkv = nn.Dense(
            self.dim * 3, use_bias=self.use_bias, kernel_init=self.kernel_init
        )(inputs)
        qkv = jnp.reshape(
            qkv, (batch, n, 3, self.num_heads, channels // self.num_heads)
        )
        qkv = jnp.transpose(qkv, (2, 0, 3, 1, 4))
        q, k, v = qkv[0], qkv[1], qkv[2]

        attention = (q @ jnp.swapaxes(k, -2, -1)) * scale
        attention = nn.softmax(attention, axis=-1)
        attention = nn.Dropout(self.att_drop)(attention, deterministic)

        x = (attention @ v).swapaxes(1, 2).reshape(batch, n, channels)
        x = nn.Dense(self.dim, kernel_init=nn.initializers.xavier_uniform())(x)
        x = nn.Dropout(self.proj_drop)(x, deterministic)

        return x



import jax.numpy as jnp

import flax.linen as nn
from ..layers import TransformerMLP
from typing import Optional

__all__ = [
    "MLPMixer",
    "MLPMixer_S16",
    "MLPMixer_S32",
    "MLPMixer_B32",
    "MLPMixer_L16",
    "MLPMixer_L32",
    "MLPMixer_H14",
]


class MixerLayer(nn.Module):
    channels_dim: int = 256
    tokens_dim: int = 256
    dropout: float = 0.2
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        norm = nn.LayerNorm()(inputs)
        transpose = jnp.swapaxes(norm, 1, 2)

        mlp1 = TransformerMLP(self.tokens_dim, transpose.shape[-1], self.dropout)(
            transpose, deterministic
        )
        transpose = jnp.swapaxes(mlp1, 1, 2)
        skip = inputs + transpose

        norm = nn.LayerNorm()(skip)
        mlp2 = TransformerMLP(self.channels_dim, norm.shape[-1], self.dropout)(
            norm, deterministic
        )
        skip = skip + mlp2

        return skip


class MLPMixer(nn.Module):
    patch_size: int = 32
    num_mixers_layers: int = 2
    hidden_size: int = 768
    channels_dim: int = 256
    tokens_dim: int = 256
    dropout: float = 0.2
    attach_head: Optional[bool] = False
    num_classes: int = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        x = nn.Conv(
            self.hidden_size,
            (self.patch_size, self.patch_size),
            strides=self.patch_size,
        )(inputs)
        batch, height, width, channels = x.shape
        x = jnp.reshape(x, (batch, height * width, channels))

        for _ in range(self.num_mixers_layers):
            x = MixerLayer(self.channels_dim, self.tokens_dim, self.dropout)(
                x, deterministic
            )

        if self.attach_head:
            x = jnp.mean(x, -1)
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x, -1)
        return x


def MLPMixer_S32(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(32, 8, 512, 2048, 256, dropout, attach_head, num_classes, **kwargs)


def MLPMixer_S16(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(16, 8, 512, 2048, 256, dropout, attach_head, num_classes, **kwargs)


def MLPMixer_B32(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(32, 12, 768, 3072, 384, dropout, attach_head, num_classes, **kwargs)


def MLPMixer_L32(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(
        32, 24, 1024, 4096, 512, dropout, attach_head, num_classes, **kwargs
    )


def MLPMixer_L16(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(
        16, 24, 1024, 4096, 512, dropout, attach_head, num_classes, **kwargs
    )


def MLPMixer_H14(attach_head=False, num_classes=1000, dropout=0.2, **kwargs):
    return MLPMixer(
        14, 32, 1280, 5120, 640, dropout, attach_head, num_classes, **kwargs
    )



from .squeeze_and_excite_layer import SqueezeAndExcitation
from .depthwise_separable_conv import DepthwiseConv2D, SeparableDepthwiseConv2D
from .drop import DropPath
from .mlp import TransformerMLP
from .attention import Attention
from .mask import Mask
from .patch_util import ExtractPatches, MergePatches, PatchEmbed
from .pool import AdaptiveAveragePool1D



import jax.numpy as jnp

import flax.linen as nn
from typing import Optional, Union, Sequence, Iterable

from ..layers import DepthwiseConv2D, SeparableDepthwiseConv2D
from ..activations import hardswish

__all__ = ["MPViT", "MPViT_Tiny", "MPViT_XSmall", "MPViT_Small", "MPViT_Base"]


class ConvolutionalStem(nn.Module):
    emb_dim: int = 128
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        conv = nn.Conv(
            self.emb_dim // 2, (3, 3), strides=2, padding="SAME", use_bias=False
        )(inputs)
        conv = nn.BatchNorm(deterministic)(conv)
        conv = hardswish(conv)

        conv = nn.Conv(self.emb_dim, (3, 3), strides=2, padding="SAME", use_bias=False)(
            conv
        )
        conv = nn.BatchNorm(deterministic)(conv)
        conv = hardswish(conv)
        b, h, w, c = conv.shape
        return jnp.reshape(conv, (b, h * w, c))


class MultiScalePatchEmbedding(nn.Module):
    features: int = 64
    kernel_size: Union[int, Sequence[int]] = (3, 3)
    strides: Union[int, Sequence[int]] = 1
    padding: Optional[Union[int, str]] = "SAME"
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        batch, dim, channels = inputs.shape
        height = width = int(dim ** 0.5)
        inputs = jnp.reshape(inputs, (batch, height, width, channels))

        conv3x3 = SeparableDepthwiseConv2D(
            self.features, 1, (3, 3), (self.strides, self.strides)
        )(inputs)
        conv3x3 = hardswish(conv3x3)
        conv3x3 = nn.BatchNorm(deterministic)(conv3x3)
        conv3x3 = jnp.reshape(conv3x3, (batch, -1, self.features))

        conv5x5 = SeparableDepthwiseConv2D(self.features, 1, (3, 3), (1, 1))(inputs)
        conv5x5 = hardswish(conv5x5)
        conv5x5 = nn.BatchNorm(deterministic)(conv5x5)
        conv5x5 = SeparableDepthwiseConv2D(
            self.features, 1, (3, 3), (self.strides, self.strides)
        )(conv5x5)
        conv5x5 = hardswish(conv5x5)
        conv5x5 = nn.BatchNorm(deterministic)(conv5x5)
        conv5x5 = jnp.reshape(conv5x5, (batch, -1, self.features))

        conv7x7 = SeparableDepthwiseConv2D(self.features, 1, (3, 3), (1, 1))(inputs)
        conv7x7 = hardswish(conv7x7)
        conv7x7 = nn.BatchNorm(deterministic)(conv7x7)
        conv7x7 = SeparableDepthwiseConv2D(self.features, 1, (3, 3), (1, 1))(conv7x7)
        conv7x7 = hardswish(conv7x7)
        conv7x7 = nn.BatchNorm(deterministic)(conv7x7)
        conv7x7 = SeparableDepthwiseConv2D(
            self.features, 1, (3, 3), (self.strides, self.strides)
        )(conv7x7)
        conv7x7 = hardswish(conv7x7)
        conv7x7 = nn.BatchNorm(deterministic)(conv7x7)
        conv7x7 = jnp.reshape(conv7x7, (batch, -1, self.features))

        return [conv3x3, conv5x5, conv7x7]


class ConvolutionalLocalFeature(nn.Module):
    features: int = 64
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        batch, n, channels = inputs.shape
        height = width = int(n ** 0.5)
        inputs = jnp.reshape(inputs, (batch, height, width, channels))

        skip = inputs
        conv = SeparableDepthwiseConv2D(self.features, 1, (1, 1), (1, 1), "SAME")(
            inputs
        )
        conv = nn.BatchNorm(use_running_average=deterministic)(conv)
        conv = hardswish(conv)

        conv = DepthwiseConv2D((3, 3), (1, 1), "SAME", 1)(conv)
        conv = nn.BatchNorm(use_running_average=deterministic)(conv)
        conv = hardswish(conv)

        conv = SeparableDepthwiseConv2D(self.features, 1, (1, 1), (1, 1), "SAME")(conv)
        conv = nn.BatchNorm(use_running_average=deterministic)(conv)
        conv = hardswish(conv)

        residual = conv + skip
        batch, height, width, channels = residual.shape
        return jnp.reshape(residual, (batch, height * width, channels))


class ConvPosEnc(nn.Module):
    """
    Implementation translated from the official repository of CoaT: https://github.com/mlpc-ucsd/CoaT
    """

    dim: int

    @nn.compact
    def __call__(self, inputs):
        batch, n, channels = inputs.shape
        height = width = int((n - 1) ** 0.5)
        cls_token, img_tokens = inputs[:, :1], inputs[:, 1:]
        features = jnp.reshape(img_tokens, [batch, height, width, channels])
        dwconv = DepthwiseConv2D((3, 3))(features)
        x = dwconv + features
        x = jnp.reshape(x, [batch, height * width, channels])

        x = jnp.concatenate([cls_token, x], axis=1)
        return x


class ConvRelPosEnc(nn.Module):
    """
    Implementation translated from the official repository of CoaT: https://github.com/mlpc-ucsd/CoaT
    """

    window_size: int

    @nn.compact
    def __call__(self, q, v, size: tuple):
        batch, num_heads, n, channels_per_head = q.shape
        height, width = size
        assert n == 1 + height * width

        windows = {self.window_size: num_heads}
        conv_list = []
        head_splits = []

        for window, head_split in windows.items():
            conv_list.append(DepthwiseConv2D((window, window), channel_multiplier=1))
            head_splits.append(head_split)
        channel_splits = [x * channels_per_head for x in head_splits]

        q_img = q[:, :, 1:, :]
        v_img = v[:, :, 1:, :]

        v_img = jnp.reshape(
            v_img, (batch, num_heads * channels_per_head, height, width)
        )
        v_img_list = jnp.array_split(v_img, channel_splits, axis=1)
        conv_v_img_list = [
            conv(jnp.transpose(x, (0, 2, 3, 1)))
            for conv, x in zip(conv_list, v_img_list)
        ]
        conv_v_img = jnp.concatenate(conv_v_img_list, axis=-1)
        conv_v_img = jnp.reshape(
            conv_v_img, (batch, num_heads, height * width, channels_per_head)
        )

        EV_hat_img = q_img * conv_v_img
        zero = jnp.zeros([batch, num_heads, 1, channels_per_head])
        EV_hat = jnp.concatenate((zero, EV_hat_img), axis=2)
        return EV_hat


class FactorizedAttention(nn.Module):
    """
    Implementation translated from the official repository of CoaT: https://github.com/mlpc-ucsd/CoaT
    """

    dim: int
    num_heads: int
    drop_prob: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        head_dim = self.dim // self.num_heads
        scale = head_dim ** -0.5

        batch, n, channels = inputs.shape
        height = width = int((n - 1) ** 0.5)
        size = (height, width)

        # Conv Positional Encoding
        pos_enc = ConvPosEnc(channels)(inputs)
        assert pos_enc.shape == inputs.shape

        # Generate Q, K, V.
        qkv = nn.Dense(self.dim * 3, use_bias=False)(pos_enc)
        qkv = qkv.reshape(batch, n, 3, self.num_heads, channels // self.num_heads)
        qkv = jnp.transpose(qkv, (2, 0, 3, 1, 4))

        q, k, v = qkv[0], qkv[1], qkv[2]

        # Factorized attention.
        k_softmax = nn.softmax(k, axis=2)
        k_softmax_T_dot_v = jnp.einsum("b h n k, b h n v -> b h k v", k_softmax, v)
        factor_att = jnp.einsum("b h n k, b h k v -> b h n v", q, k_softmax_T_dot_v)

        # ConvRel Positional Encoding
        crpe = ConvRelPosEnc(1)(q, v, size)

        # Merge and reshape.
        x = scale * factor_att + crpe
        x = x.transpose(0, 1, 2, 3).reshape(batch, n, channels)

        # Output projection.
        x = nn.Dense(self.dim)(x)
        x = nn.Dropout(self.drop_prob)(x, deterministic)

        return x


class TransformerEncoder(nn.Module):
    dim: int = 256
    num_heads: int = 2
    att_drop: float = 0.2
    proj_drop: float = 0.2
    mlp_ratio: int = 2
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        skip = inputs
        norm = nn.LayerNorm()(inputs)
        mhsa = FactorizedAttention(
            self.dim, self.num_heads, self.att_drop, deterministic
        )(norm)
        mhsa = mhsa + skip
        norm = nn.LayerNorm()(mhsa)
        x = nn.Dense(self.dim * self.mlp_ratio)(norm)
        x = nn.gelu(x)
        x = nn.Dropout(self.proj_drop)(x, deterministic=deterministic)
        x = nn.Dense(self.dim)(x)
        x = x + mhsa
        return x


class MultiPathTransformerBlock(nn.Module):
    deterministic: Optional[bool] = None
    features: int = 64
    dim: int = 64
    num_heads: int = 2
    att_drop: float = 0.2
    proj_drop: float = 0.2
    mlp_ratio: int = 2
    num_encoder_layers: int = 1

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        x1, x2, x3 = inputs
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        clf = ConvolutionalLocalFeature(self.features)(x1, deterministic=deterministic)

        for index in range(self.num_encoder_layers):
            batch, n, channels = x1.shape

            # Add cls token
            x1 = jnp.concatenate(
                [
                    self.param(
                        f"param_1_{index}", nn.initializers.zeros, (batch, 1, channels)
                    ),
                    x1,
                ],
                axis=1,
            )
            x2 = jnp.concatenate(
                [
                    self.param(
                        f"param_2_{index}", nn.initializers.zeros, (batch, 1, channels)
                    ),
                    x2,
                ],
                axis=1,
            )
            x3 = jnp.concatenate(
                [
                    self.param(
                        f"param_3_{index}", nn.initializers.zeros, (batch, 1, channels)
                    ),
                    x3,
                ],
                axis=1,
            )

            x1 = TransformerEncoder(
                self.dim, self.num_heads, self.att_drop, self.proj_drop, self.mlp_ratio
            )(x1, deterministic)
            x2 = TransformerEncoder(
                self.dim, self.num_heads, self.att_drop, self.proj_drop, self.mlp_ratio
            )(x2, deterministic)
            x3 = TransformerEncoder(
                self.dim, self.num_heads, self.att_drop, self.proj_drop, self.mlp_ratio
            )(x3, deterministic)

            # Remove cls token
            x1, x2, x3 = x1[:, 1:, :], x2[:, 1:, :], x3[:, 1:, :]

        concat = jnp.concatenate([clf, x1, x2, x3], axis=-1)
        conv = nn.Conv(self.features, [1, 1], 1)(concat)
        return conv


class MPViT(nn.Module):
    mlp_ratio: int = 2
    channels_list: Iterable = (64, 96, 176, 216)
    num_layers_list: Iterable = (1, 2, 4, 1)
    att_drop: float = 0.2
    proj_drop: float = 0.2
    attach_head: bool = True
    num_classes: int = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        stem = ConvolutionalStem(64)(inputs, deterministic=deterministic)
        mptb = stem
        for i, j in zip(self.channels_list, self.num_layers_list):
            mspe = MultiScalePatchEmbedding(features=i, strides=2)(mptb, deterministic)
            mptb = MultiPathTransformerBlock(
                features=i,
                dim=i,
                num_heads=8,
                num_encoder_layers=j,
                mlp_ratio=self.mlp_ratio,
                att_drop=self.att_drop,
                proj_drop=self.proj_drop,
            )(mspe, deterministic)

        if self.attach_head:
            # Global avg pooling
            x = jnp.mean(mptb, 1)
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)
            return x

        return mptb


def MPViT_Tiny(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    return MPViT(
        mlp_ratio=2,
        channels_list=(64, 96, 176, 216),
        num_layers_list=(1, 2, 4, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        att_drop=dropout,
        proj_drop=dropout,
        **kwargs,
    )


def MPViT_XSmall(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    return MPViT(
        mlp_ratio=4,
        channels_list=(64, 128, 192, 256),
        num_layers_list=(1, 2, 4, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        att_drop=dropout,
        proj_drop=dropout,
        **kwargs,
    )


def MPViT_Small(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    return MPViT(
        mlp_ratio=4,
        channels_list=(64, 128, 216, 288),
        num_layers_list=(1, 3, 6, 3),
        attach_head=attach_head,
        num_classes=num_classes,
        att_drop=dropout,
        proj_drop=dropout,
        **kwargs,
    )


def MPViT_Base(attach_head=True, num_classes=1000, dropout=0.1, **kwargs):
    return MPViT(
        mlp_ratio=4,
        channels_list=(128, 224, 368, 480),
        num_layers_list=(1, 3, 8, 3),
        attach_head=attach_head,
        num_classes=num_classes,
        att_drop=dropout,
        proj_drop=dropout,
        **kwargs,
    )



from . import poolformer
from . import patchconvnet
from . import mpvit
from . import mlp_mixer
from . import conv_mixer
from . import segformer
from . import masked_autoencoder
from .model_registry import list_models, load_model



import jax.numpy as jnp
from jax.image import resize
import flax.linen as nn

from typing import Optional, Iterable, Any

from ..layers import DropPath
from ..layers import TransformerMLP

__all__ = [
    "SegFormer",
    "SegFormer_B0",
    "SegFormer_B1",
    "SegFormer_B2",
    "SegFormer_B3",
    "SegFormer_B4",
    "SegFormer_B5",
]


class EfficientSelfAttention(nn.Module):
    dim: int = 256
    num_heads: int = 8
    sr_ratio: int = 1
    att_drop: float = 0.1
    proj_drop: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        assert self.dim % self.num_heads == 0

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        head_dim = self.dim // self.num_heads
        scale = head_dim ** -0.5

        batch, n, channels = inputs.shape
        height = width = int(n ** 0.5)

        q = nn.Dense(self.dim, use_bias=True)(inputs)
        q = jnp.reshape(q, (batch, n, self.num_heads, channels // self.num_heads))
        q = jnp.transpose(q, (0, 2, 1, 3))

        if self.sr_ratio > 1:
            inputs = jnp.transpose(inputs, (0, 2, 1)).reshape(
                batch, height, width, channels
            )
            conv = nn.Conv(self.dim, (self.sr_ratio, self.sr_ratio), self.sr_ratio)(
                inputs
            )
            conv = jnp.reshape(conv, (batch, -1, channels))
            norm = nn.LayerNorm()(conv)
            kv = nn.Dense(self.dim * 2, use_bias=True)(norm).reshape(
                batch, -1, 2, self.num_heads, channels // self.num_heads
            )
            kv = jnp.transpose(kv, (2, 0, 3, 1, 4))

        else:
            kv = nn.Dense(self.dim * 2, use_bias=True)(inputs)
            kv = jnp.reshape(
                kv, (batch, -1, 2, self.num_heads, channels // self.num_heads)
            )
            kv = jnp.transpose(kv, (2, 0, 3, 1, 4))

        k, v = kv[0], kv[1]

        att = q @ jnp.swapaxes(k, -2, -1) * scale
        att = nn.softmax(att, -1)
        att = nn.Dropout(self.att_drop)(att, deterministic)

        x = jnp.swapaxes(att @ v, 1, 2).reshape(batch, n, channels)
        x = nn.Dense(self.dim)(x)
        x = nn.Dropout(self.proj_drop)(x, deterministic)

        return x


class Block(nn.Module):
    dim: int = 256
    num_heads: int = 8
    mlp_ratio: int = 4
    sr_ratio: int = 1
    att_drop: float = 0.1
    drop: float = 0.1
    drop_path: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        x = nn.LayerNorm()(inputs)
        x = EfficientSelfAttention(
            self.dim, self.num_heads, self.sr_ratio, self.att_drop, self.drop
        )(x, deterministic)
        x = DropPath(self.drop_path)(x, deterministic)
        inputs = inputs + x

        x = nn.LayerNorm()(inputs)
        x = TransformerMLP(
            self.dim * self.mlp_ratio, self.dim, self.drop, use_dwconv=True
        )(x, deterministic)
        x = DropPath(self.drop_path)(x, deterministic)
        return inputs + x


class OverlapPatchEmbed(nn.Module):
    emb_dim: int = 768
    patch_size: int = 16
    stride: int = 4

    @nn.compact
    def __call__(self, inputs):
        conv = nn.Conv(self.emb_dim, (self.patch_size, self.patch_size), self.stride)(
            inputs
        )
        flat = jnp.reshape(conv, (conv.shape[0], -1, conv.shape[-1]))
        norm = nn.LayerNorm()(flat)
        return norm


class MixTransformer(nn.Module):
    patch_size: int = 16
    emb_dims: Iterable = (64, 128, 256, 512)
    num_heads: Iterable = (1, 2, 4, 8)
    mlp_ratios = (4, 4, 4, 4)
    drop: float = 0.1
    att_drop: float = 0.1
    drop_path: float = 0.1
    depths: Iterable = (3, 4, 6, 3)
    sr_ratios: Iterable = (8, 4, 2, 1)
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        batch = inputs.shape[0]
        outs = []

        dpr = [x.item() for x in jnp.linspace(0, self.drop_path, sum(self.depths))]
        cur = 0

        x = OverlapPatchEmbed(self.emb_dims[0], 7, 4)(inputs)
        for i in range(self.depths[0]):
            x = Block(
                self.emb_dims[0],
                self.num_heads[0],
                self.mlp_ratios[0],
                self.sr_ratios[0],
                self.att_drop,
                self.drop,
                dpr[cur + i],
            )(x, deterministic)
        x = nn.LayerNorm()(x)
        height = width = int(x.shape[1] ** 0.5)
        x = jnp.reshape(x, (batch, height, width, -1))
        outs.append(x)

        cur += self.depths[0]
        x = OverlapPatchEmbed(self.emb_dims[1], 3, 2)(x)
        for i in range(self.depths[1]):
            x = Block(
                self.emb_dims[1],
                self.num_heads[1],
                self.mlp_ratios[1],
                self.sr_ratios[1],
                self.att_drop,
                self.drop,
                dpr[cur + i],
            )(x, deterministic)
        x = nn.LayerNorm()(x)
        height = width = int(x.shape[1] ** 0.5)
        x = jnp.reshape(x, (batch, height, width, -1))
        outs.append(x)

        cur += self.depths[1]
        x = OverlapPatchEmbed(self.emb_dims[2], 3, 2)(x)
        for i in range(self.depths[2]):
            x = Block(
                self.emb_dims[2],
                self.num_heads[2],
                self.mlp_ratios[2],
                self.sr_ratios[2],
                self.att_drop,
                self.drop,
                dpr[cur + i],
            )(x, deterministic)
        x = nn.LayerNorm()(x)
        height = width = int(x.shape[1] ** 0.5)
        x = jnp.reshape(x, (batch, height, width, -1))
        outs.append(x)

        cur += self.depths[2]
        x = OverlapPatchEmbed(self.emb_dims[3], 3, 2)(x)
        for i in range(self.depths[3]):
            x = Block(
                self.emb_dims[3],
                self.num_heads[3],
                self.mlp_ratios[3],
                self.sr_ratios[3],
                self.att_drop,
                self.drop,
                dpr[cur + i],
            )(x, deterministic)
        x = nn.LayerNorm()(x)
        height = width = int(x.shape[1] ** 0.5)
        x = jnp.reshape(x, (batch, height, width, -1))
        outs.append(x)

        return outs


class MLPHead(nn.Module):
    emb_dim: int = 768

    @nn.compact
    def __call__(self, inputs):
        batch, height, width, channels = inputs.shape
        x = jnp.reshape(inputs, (batch, height * width, channels))
        x = nn.Dense(self.emb_dim)(x)
        return x


class SegFormerHead(nn.Module):
    emb_dim: int = 768
    kernel_size: Iterable[int] = (1, 1)
    stride: int = 1
    num_classes: int = 3
    drop: float = 0.1
    axis_name: Optional[str] = None
    axis_index_groups: Any = None
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        c1, c2, c3, c4 = inputs
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        h1 = MLPHead(self.emb_dim)(c1)
        h1 = jnp.reshape(h1, (c1.shape[0], c1.shape[1], c1.shape[2], -1))

        h2 = MLPHead(self.emb_dim)(c2)
        h2 = jnp.reshape(h2, (c2.shape[0], c2.shape[1], c2.shape[2], -1))
        h2 = resize(
            h2, (c1.shape[0], c1.shape[1], c1.shape[2], h2.shape[-1]), "bilinear"
        )

        h3 = MLPHead(self.emb_dim)(c3)
        h3 = jnp.reshape(h3, (c3.shape[0], c3.shape[1], c3.shape[2], -1))
        h3 = resize(
            h3, (c1.shape[0], c1.shape[1], c1.shape[2], h3.shape[-1]), "bilinear"
        )

        h4 = MLPHead(self.emb_dim)(c4)
        h4 = jnp.reshape(h4, (c4.shape[0], c4.shape[1], c4.shape[2], -1))
        h4 = resize(
            h4, (c1.shape[0], c1.shape[1], c1.shape[2], h4.shape[-1]), "bilinear"
        )

        concat = jnp.concatenate([h1, h2, h3, h4], axis=-1)

        conv = nn.Conv(self.emb_dim, self.kernel_size, self.stride)(concat)
        norm = nn.BatchNorm(
            deterministic,
            axis_name=self.axis_name,
            axis_index_groups=self.axis_index_groups,
        )(conv)
        act = nn.relu(norm)

        dropout = nn.Dropout(self.drop)(act, deterministic)
        linear_pred = nn.Conv(self.num_classes, (1, 1))(dropout)

        return linear_pred


class SegFormer(nn.Module):
    patch_size: int = 4
    emb_dims: Iterable = (64, 128, 256, 512)
    num_heads: Iterable = (1, 2, 4, 8)
    mlp_ratios = (4, 4, 4, 4)
    drop: float = 0.1
    att_drop: float = 0.1
    drop_path: float = 0.1
    depths: Iterable = (3, 4, 6, 3)
    sr_ratios: Iterable = (8, 4, 2, 1)
    num_classes: int = 19
    decoder_emb: int = 256
    axis_name: Optional[str] = None
    axis_index_groups: Any = None
    attach_head: bool = False
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        out_list = MixTransformer(
            self.patch_size,
            self.emb_dims,
            self.num_heads,
            self.drop,
            self.drop,
            self.drop_path,
            self.depths,
            self.sr_ratios,
        )(inputs, deterministic)
        if self.attach_head:
            return SegFormerHead(
                emb_dim=self.decoder_emb,
                drop=self.drop,
                num_classes=self.num_classes,
                axis_name=self.axis_name,
                axis_index_groups=self.axis_index_groups,
            )(out_list, deterministic)
        else:
            return out_list


def SegFormer_B0(
    attach_head=False,
    num_classes=19,
    dropout=0.0,
    drop_path=0.1,
    **kwargs,
):
    return SegFormer(
        4,
        (32, 64, 160, 256),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (2, 2, 2, 2),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=256,
        **kwargs,
    )


def SegFormer_B1(
    attach_head=False, num_classes=19, dropout=0.0, drop_path=0.1, **kwargs
):
    return SegFormer(
        4,
        (64, 128, 320, 512),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (2, 2, 2, 2),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=256,
        **kwargs,
    )


def SegFormer_B2(
    attach_head=False, num_classes=19, dropout=0.0, drop_path=0.1, **kwargs
):
    return SegFormer(
        4,
        (64, 128, 320, 512),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (3, 4, 6, 3),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=768,
        **kwargs,
    )


def SegFormer_B3(
    attach_head=False, num_classes=19, dropout=0.0, drop_path=0.1, **kwargs
):
    return SegFormer(
        4,
        (64, 128, 320, 512),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (3, 4, 18, 3),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=768,
        **kwargs,
    )


def SegFormer_B4(
    attach_head=False, num_classes=19, dropout=0.0, drop_path=0.1, **kwargs
):
    return SegFormer(
        4,
        (64, 128, 320, 512),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (3, 8, 27, 3),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=768,
        **kwargs,
    )


def SegFormer_B5(
    attach_head=False, num_classes=19, dropout=0.0, drop_path=0.1, **kwargs
):
    return SegFormer(
        4,
        (64, 128, 320, 512),
        (1, 2, 5, 8),
        dropout,
        dropout,
        drop_path,
        (3, 6, 40, 3),
        (8, 4, 2, 1),
        attach_head=attach_head,
        num_classes=num_classes,
        decoder_emb=768,
        **kwargs,
    )



import jax.numpy as jnp
import flax.linen as nn

from typing import Union, Iterable


class AdaptiveAveragePool1D(nn.Module):
    output_size: Union[int, Iterable[int]]

    @nn.compact
    def __call__(self, inputs):
        output_size = (
            (self.output_size,)
            if isinstance(self.output_size, int)
            else self.output_size
        )
        split = jnp.split(inputs, output_size[0], axis=1)
        stack = jnp.stack(split, axis=1)
        return jnp.mean(stack, axis=2)



import jax.numpy as jnp
import flax.linen as nn

from typing import Optional
from ..layers import DepthwiseConv2D

__all__ = [
    "ConvMixer",
    "ConvMixer_512_12",
    "ConvMixer_768_32",
    "ConvMixer_1024_20",
    "ConvMixer_1536_20",
]


class ConvMixerLayer(nn.Module):
    features: int = 256
    filter_size: int = 3
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        x = inputs
        x = DepthwiseConv2D()(x)
        x = nn.gelu(x)
        x = nn.BatchNorm(deterministic)(x)

        add = x + inputs

        x = nn.Conv(self.features, (self.filter_size, self.filter_size))(add)
        x = nn.gelu(x)
        x = nn.BatchNorm(deterministic)(x)

        return x


class ConvMixer(nn.Module):
    features: int = 512  # h
    patch_size: int = 7
    num_mixer_layers: int = 12  # depth
    filter_size: int = 9
    attach_head: bool = False
    num_classes: int = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):
        extract_patches = nn.Conv(
            self.features, (self.patch_size, self.patch_size), self.patch_size
        )(inputs)
        x = nn.gelu(extract_patches)
        x = nn.BatchNorm(deterministic)(x)
        for _ in range(self.num_mixer_layers):
            x = ConvMixerLayer(self.features, self.filter_size)(x, deterministic)

        if self.attach_head:
            x = jnp.mean(x, [1, 2])
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


def ConvMixer_1536_20(attach_head=False, num_classes=1000, dropout=None, **kwargs):
    return ConvMixer(1536, 7, 20, 9, attach_head, num_classes, **kwargs)


def ConvMixer_768_32(attach_head=False, num_classes=1000, dropout=None, **kwargs):
    return ConvMixer(768, 7, 32, 7, attach_head, num_classes, **kwargs)


def ConvMixer_512_12(attach_head=False, num_classes=1000, dropout=None, **kwargs):
    return ConvMixer(512, 7, 12, 8, attach_head, num_classes, **kwargs)


def ConvMixer_1024_20(attach_head=False, num_classes=1000, dropout=None, **kwargs):
    return ConvMixer(1024, 14, 20, 9, attach_head, num_classes, **kwargs)



"""
A major part of this code is translated from https://github.com/microsoft/Swin-Transformer
"""

import jax.numpy as jnp
import flax.linen as nn

from ..layers import PatchEmbed, TransformerMLP, DropPath, AdaptiveAveragePool1D

from typing import Iterable, Optional

__all__ = [
    "SwinTransformer",
    "SwinTiny224",
    "SwinSmall224",
    "SwinBase224",
    "SwinBase384",
    "SwinLarge224",
    "SwinLarge384",
]


def window_partition(x, window_size):
    batch, height, width, channels = x.shape
    x = jnp.reshape(
        x,
        (
            batch,
            height // window_size,
            window_size,
            width // window_size,
            window_size,
            channels,
        ),
    )
    windows = jnp.reshape(
        jnp.transpose(x, (0, 1, 3, 2, 4, 5)), (-1, window_size, window_size, channels)
    )
    return windows


def window_reverse(windows, window_size, height, width):
    batch = int(windows.shape[0] / (height * width / window_size / window_size))
    x = jnp.reshape(
        windows,
        (
            batch,
            height // window_size,
            width // window_size,
            window_size,
            window_size,
            -1,
        ),
    )
    x = jnp.reshape(jnp.transpose(x, (0, 1, 3, 2, 4, 5)), (batch, height, width, -1))
    return x


class WindowAttention(nn.Module):
    dim: int
    num_heads: int
    window_size: Iterable[int]
    use_bias: bool = True
    att_drop: float = 0.0
    proj_drop: float = 0.0
    deterministic: Optional[bool] = None

    def get_rel_pos_index(self):
        h_indices = jnp.arange(0, self.window_size[0])
        w_indices = jnp.arange(0, self.window_size[1])
        indices = jnp.stack(jnp.meshgrid(h_indices, w_indices))
        flatten_indices = jnp.reshape(indices, (2, -1))
        relative_indices = flatten_indices[:, :, None] - flatten_indices[:, None, :]
        relative_indices = jnp.transpose(relative_indices, (1, 2, 0))

        relative_indices.at[:, :, 0].add(self.window_size[0] - 1)
        relative_indices.at[:, :, 1].add(self.window_size[1] - 1)
        relative_indices.at[:, :, 0].mul(2 * self.window_size[1] - 1)
        relative_pos_index = jnp.sum(relative_indices, -1)
        return relative_pos_index

    @nn.compact
    def __call__(self, inputs, mask=None, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        rpbt = self.param(
            "rel_pos_bias_table",
            nn.initializers.normal(0.02),
            (
                (2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1),
                self.num_heads,
            ),
        )

        relative_pos_index = self.variable(
            "rel_pos_index", "rel_pos_index", self.get_rel_pos_index
        )

        batch, n, channels = inputs.shape
        qkv = nn.Dense(self.dim * 3, use_bias=self.use_bias)(inputs)
        qkv = qkv.reshape(batch, n, 3, self.num_heads, channels // self.num_heads)
        qkv = jnp.transpose(qkv, (2, 0, 3, 1, 4))
        q, k, v = qkv[0], qkv[1], qkv[2]

        q = q * ((self.dim // self.num_heads) ** -0.5)
        att = q @ jnp.swapaxes(k, -2, -1)

        rel_pos_bias = jnp.reshape(
            rpbt[jnp.reshape(relative_pos_index.value, (-1))],
            (
                self.window_size[0] * self.window_size[1],
                self.window_size[0] * self.window_size[1],
                -1,
            ),
        )
        rel_pos_bias = jnp.transpose(rel_pos_bias, (2, 0, 1))
        att += jnp.expand_dims(rel_pos_bias, 0)

        if mask is not None:
            att = jnp.reshape(
                att, (batch // mask.shape[0], mask.shape[0], self.num_heads, n, n)
            )
            att = att + jnp.expand_dims(jnp.expand_dims(mask, 1), 0)
            att = jnp.reshape(att, (-1, self.num_heads, n, n))
            att = nn.softmax(att)

        else:
            att = nn.softmax(att)

        att = nn.Dropout(self.att_drop)(att, deterministic)

        x = jnp.reshape(jnp.swapaxes(att @ v, 1, 2), (batch, n, channels))
        x = nn.Dense(self.dim)(x)
        x = nn.Dropout(self.proj_drop)(x, deterministic)
        return x


class SwinBlock(nn.Module):
    dim: int
    inp_hw: Iterable[int]
    num_heads: int
    window_size: int = 7
    shift_size: int = 0
    mlp_ratio: int = 4
    use_att_bias: bool = True
    dropout: float = 0.0
    att_dropout: float = 0.0
    drop_path: float = 0.0
    deterministic: Optional[bool] = None

    def get_att_mask(self, shift_size, window_size, height, width):
        if shift_size > 0:
            mask = jnp.zeros([1, height, width, 1])
            h_slices = (
                slice(0, -window_size),
                slice(-window_size, -shift_size),
                slice(-shift_size, None),
            )
            w_slices = (
                slice(0, -window_size),
                slice(-window_size, -shift_size),
                slice(-shift_size, None),
            )

            count = 0
            for h in h_slices:
                for w in w_slices:
                    mask.at[:, h, w, :].set(count)
                    count += 1

            mask_windows = window_partition(mask, window_size)
            mask_windows = jnp.reshape(mask_windows, (-1, window_size * window_size))
            att_mask = jnp.expand_dims(mask_windows, 1) - jnp.expand_dims(
                mask_windows, 2
            )
            att_mask = jnp.where(att_mask != 0, att_mask, float(-100.0))
            att_mask = jnp.where(att_mask == 0, att_mask, float(0.0))

        else:
            att_mask = None
            return att_mask

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        height, width = self.inp_hw[0], self.inp_hw[1]

        input_resolution = min(height, width)
        if input_resolution <= self.window_size:
            shift_size = 0
            window_size = input_resolution
        else:
            shift_size = self.shift_size
            window_size = self.window_size

        att_mask = self.variable(
            "attention_mask",
            "mask",
            self.get_att_mask,
            shift_size,
            window_size,
            height,
            width,
        )

        batch, length, channels = inputs.shape
        height, width = self.inp_hw[0], self.inp_hw[1]
        assert length == height * width

        residual = inputs
        x = nn.LayerNorm()(inputs)
        x = jnp.reshape(x, (batch, height, width, channels))

        if shift_size > 0:
            shifted_x = jnp.roll(x, (-shift_size, -shift_size), axis=(1, 2))
        else:
            shifted_x = x

        x_windows = window_partition(shifted_x, window_size)
        x_windows = jnp.reshape(x_windows, (-1, window_size * window_size, channels))

        att_windows = WindowAttention(
            self.dim,
            self.num_heads,
            (window_size, window_size),
            self.use_att_bias,
            self.att_dropout,
            self.dropout,
        )(x_windows, att_mask.value, deterministic)

        att_windows = jnp.reshape(att_windows, (-1, window_size, window_size, channels))
        shifted_x = window_reverse(att_windows, window_size, height, width)

        if shift_size > 0:
            x = jnp.roll(shifted_x, (shift_size, shift_size), axis=(1, 2))
        else:
            x = shifted_x

        x = jnp.reshape(x, (batch, height * width, channels))

        x = residual + DropPath(self.drop_path)(x, deterministic)
        x = nn.LayerNorm()(x)
        mlp = TransformerMLP(self.dim * self.mlp_ratio, self.dim)(x, deterministic)
        x = x + DropPath(self.drop_path)(mlp, deterministic)

        return x


class PatchMerging(nn.Module):
    inp_res: Iterable[int]
    dim: int

    @nn.compact
    def __call__(self, inputs):
        batch, n, channels = inputs.shape
        height, width = self.inp_res[0], self.inp_res[1]
        x = jnp.reshape(inputs, (batch, height, width, channels))

        x0 = x[:, 0::2, 0::2, :]
        x1 = x[:, 1::2, 0::2, :]
        x2 = x[:, 0::2, 1::2, :]
        x3 = x[:, 1::2, 1::2, :]

        x = jnp.concatenate([x0, x1, x2, x3], axis=-1)
        x = jnp.reshape(x, (batch, -1, 4 * channels))
        x = nn.LayerNorm()(x)
        x = nn.Dense(2 * self.dim, use_bias=False)(x)
        return x


class SwinLayer(nn.Module):
    inp_res: Iterable[int]
    dim: int
    num_heads: int
    window_size: int
    mlp_ratio: int
    use_att_bias: bool
    dropout: float
    att_dropout: float
    drop_path: float
    depth: int
    downsample: Optional[bool] = False
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, x, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        for i in range(self.depth):
            x = SwinBlock(
                self.dim,
                self.inp_res,
                self.num_heads,
                self.window_size,
                0 if (i % 2 == 0) else self.window_size // 2,
                self.mlp_ratio,
                self.use_att_bias,
                self.dropout,
                self.att_dropout,
                self.drop_path[i]
                if isinstance(self.drop_path, (list, tuple))
                else self.drop_path,
            )(x, deterministic)

        if self.downsample:
            x = PatchMerging(self.inp_res, self.dim)(x)

        return x


class SwinTransformer(nn.Module):
    patch_size: int = 4
    emb_dim: int = 96
    depths: Iterable[int] = (2, 2, 6, 2)
    num_heads: Iterable[int] = (3, 6, 12, 24)
    window_size: int = 7
    mlp_ratio: int = 4
    use_att_bias: bool = True
    dropout: float = 0.0
    att_dropout: float = 0.0
    drop_path: float = 0.1
    use_abs_pos_emb: bool = False
    attach_head: bool = True
    num_classes: Optional[int] = 1000
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, inputs, deterministic=None):

        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )

        x = PatchEmbed(self.patch_size, self.emb_dim, use_norm=True)(inputs)
        num_patches = x.shape[1]
        patch_grid = (
            inputs.shape[1] // self.patch_size,
            inputs.shape[2] // self.patch_size,
        )

        if self.use_abs_pos_emb:
            abs_pos_emb = self.param(
                "abs_pos_emb",
                nn.initializers.normal(0.02),
                (1, num_patches, self.emb_dim),
            )
            x = x + abs_pos_emb

        x = nn.Dropout(self.dropout)(x, deterministic)

        dpr = [x.item() for x in jnp.linspace(0, self.drop_path, sum(self.depths))]

        layers = []
        for i in range(len(self.depths)):
            x = SwinLayer(
                (patch_grid[0] // (2 ** i), patch_grid[1] // (2 ** i)),
                int(self.emb_dim * 2 ** i),
                self.num_heads[i],
                self.window_size,
                self.mlp_ratio,
                self.use_att_bias,
                self.dropout,
                self.att_dropout,
                self.drop_path,
                self.depths[i],
                True if i < (len(self.depths) - 1) else False,
            )(x, deterministic)
        x = nn.LayerNorm()(x)
        x = AdaptiveAveragePool1D(1)(x)
        x = jnp.reshape(x, (1, -1))

        if self.attach_head:
            x = nn.Dense(self.num_classes)(x)
            x = nn.softmax(x)

        return x


def SwinTiny224(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=96,
        depths=(2, 2, 6, 2),
        num_heads=(3, 6, 12, 24),
        window_size=7,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )


def SwinSmall224(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=96,
        depths=(2, 2, 18, 2),
        num_heads=(3, 6, 12, 24),
        window_size=7,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )


def SwinBase224(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=128,
        depths=(2, 2, 18, 2),
        num_heads=(4, 8, 16, 32),
        window_size=7,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )


def SwinLarge224(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=192,
        depths=(2, 2, 18, 2),
        num_heads=(6, 12, 24, 48),
        window_size=7,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )


def SwinBase384(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=128,
        depths=(2, 2, 18, 2),
        num_heads=(4, 8, 16, 32),
        window_size=12,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )


def SwinLarge384(attach_head=True, num_classes=1000, dropout=0.0):
    return SwinTransformer(
        patch_size=4,
        emb_dim=192,
        depths=(2, 2, 18, 2),
        num_heads=(6, 12, 24, 48),
        window_size=12,
        mlp_ratio=4,
        use_att_bias=True,
        dropout=dropout,
        att_dropout=0.0,
        drop_path=0.1,
        use_abs_pos_emb=False,
        attach_head=attach_head,
        num_classes=num_classes,
    )



import jax.numpy as jnp
import jax.random as random

import flax.linen as nn


class Mask(nn.Module):
    def __call__(self, x, mask_ratio):
        batch, length, dim = x.shape
        keep_len = int(length * (1 - mask_ratio))

        rng = self.make_rng("noise")
        noise = random.uniform(rng, (batch, length))
        ids_shuffle = jnp.argsort(noise, axis=1)
        ids_restore = jnp.argsort(ids_shuffle, axis=1)

        ids_keep = ids_shuffle[:, :keep_len]
        expanded = jnp.expand_dims(ids_keep, -1)

        repeat = jnp.tile(expanded, (1, 1, dim))
        masked = jnp.take_along_axis(x, repeat, axis=1)

        mask = jnp.ones([batch, length])
        mask.at[:, :keep_len].set(0)
        mask = jnp.take_along_axis(mask, ids_restore, axis=1)

        return masked, mask, ids_restore



import jax.numpy as jnp
import jax.random as random
import flax.linen as nn
from typing import Optional


class DropPath(nn.Module):
    """
    Implementation referred from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/drop.py
    """

    dropout_prob: float = 0.1
    deterministic: Optional[bool] = None

    @nn.compact
    def __call__(self, input, deterministic=None):
        deterministic = nn.merge_param(
            "deterministic", self.deterministic, deterministic
        )
        if deterministic:
            return input
        keep_prob = 1 - self.dropout_prob
        shape = (input.shape[0],) + (1,) * (input.ndim - 1)
        rng = self.make_rng("drop_path")
        random_tensor = keep_prob + random.uniform(rng, shape)
        random_tensor = jnp.floor(random_tensor)
        return jnp.divide(input, keep_prob) * random_tensor



from setuptools import setup
from codecs import open
from os import path

current_path = path.abspath(path.dirname(__file__))

with open(path.join(current_path, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="jax_models",
    version="0.0.3",
    description="Unofficial JAX implementations of deep learning research papers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/darshandeshpande/jax-models",
    author="Darshan Deshpande",
    author_email="darshan.g.deshpande@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="jax flax models computer-vision deep-learning transformers",
    packages=["jax_models"],
    package_data={"jax_models": ["*/*"]},
    include_package_data=True,
    install_requires=["flax >= 0.3.6"],
    python_requires=">=3.6",
)
